# Copyright (c) 2010-2011 Outpost Embedded, LLC
# This file is part of the pydbusclone software package.
#
# pydbusclone is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 2 as published by the
# Free Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from threading import RLock

import dbus

from dbusclone.base import (
  PropertiesObject,
  DBusObject,
  DBusObjectList,
)
from dbusclone.util import CallbackList


NONEXISTENT = object()


def match_connman_error(e, names):
    dbus_name = e.get_dbus_name()
    for name in names:
        if dbus_name == ('net.connman.Error.%s' % name):
            return True
    return False


class ConnmanPropertiesObject(PropertiesObject, DBusObject):
    _properties = None
    _lock = None

    def __init__(self, **kwargs):
        self._lock = RLock()

        super(ConnmanPropertiesObject, self).__init__(**kwargs)

        self.connect_to_signal(
          'PropertyChanged',
          self._handle_property_changed,
        )
        self.call(
          'GetProperties',
          reply_handler = self._handle_get_properties_reply,
        )

    def get_proxy_object(self):
        return self.parent.get_proxy_object()

    def is_ready(self):
        if not super(ConnmanPropertiesObject, self).is_ready():
            return False
        return (self._properties is not None)

    def _handle_get_properties_reply(self, properties):
        self._properties = properties
        self.maybe_ready()

    def _handle_property_changed(self, name, value):
        # Note: There is currently no mechanism by which we can be notified
        # that a property has been cleared (removed).  I believe this can
        # only happen with the Passphrase service property, for which
        # PropertyChanged signals are not sent.
        if self._properties is not None:
            with self._lock:
                if value == self._properties.get(name, NONEXISTENT):
                    changed = False
                else:
                    self._properties[name] = value
                    changed = True
                    properties = dict(self._properties)

            if changed:
                self.changed_callbacks.call(
                  self.parent,
                  [name],
                  properties,
                )

    ###

    def getdict(self, names = None):
        with self._lock:
            if names is None:
                return dict(self._properties)
            else:
                return dict((name, self._properties[name]) for name in names)

    def get(self, name, default = None):
        with self._lock:
            if default is not None:
                return self._properties.get(name, default)
            return self._properties[name]

    def set(self, name, value, reply_handler, error_handler = None):
        kwargs = {'reply_handler': reply_handler}
        if error_handler is not None:
            kwargs['error_handler'] = error_handler
        self.call('SetProperty', name, value, **kwargs)

    ###

    def refresh(self):
        self.call(
          'GetProperties',
          reply_handler = self._handle_get_properties_reply_refresh,
        )

    def _handle_get_properties_reply_refresh(self, properties):
        with self._lock:
            changed_properties = []
            for k, v in properties.items():
                if v != self._properties.get(k, NONEXISTENT):
                    changed_properties.append(k)
            for k in self._properties:
                if k not in properties:
                    changed_properties.append(k)
            self._properties = dict(properties)

        if changed_properties:
            self.changed_callbacks.call(
              self.parent,
              changed_properties,
              properties,
            )


class DBusObjectWithConnmanProperties(DBusObject):
    properties = None

    def __init__(self, **kwargs):
        super(DBusObjectWithConnmanProperties, self).__init__(**kwargs)

        self.properties = ConnmanPropertiesObject(
          bus = self.bus,
          parent = self,
          bus_name = self.bus_name,
          object_path = self.object_path,
          ready_callback = self._properties_ready,
          default_error_handler = self.default_error_handler,
        )

    def is_ready(self):
        if not super(DBusObjectWithConnmanProperties, self).is_ready():
            return False
        return (self.properties is not None) and self.properties.is_ready()

    def _properties_ready(self, properties):
        self.maybe_ready()


class ConnmanManager(DBusObjectWithConnmanProperties):
    technologies = None
    services = None

    def __init__(self, **kwargs):
        super(ConnmanManager, self).__init__(
          bus_name = 'net.connman',
          object_path = '/',
          **kwargs
        )

        self.technologies = DBusObjectList(
          bus = self.bus,
          object_cls = ConnmanTechnology,
          ready_callback = self._handle_technologies_ready,
          default_error_handler = self.default_error_handler,
        )
        self.connect_to_signal(
          'TechnologyAdded',
          (lambda path, props: self.technologies.add_object(path)),
        )
        self.connect_to_signal(
          'TechnologyRemoved',
          self.technologies.remove_object,
        )
        self.call(
          'GetTechnologies',
          reply_handler = self._handle_get_technologies_reply,
        )

        self.services = DBusObjectList(
          bus = self.bus,
          object_cls = ConnmanService,
          ready_callback = self._handle_services_ready,
          default_error_handler = self.default_error_handler,
        )
        self.connect_to_signal(
          'ServicesChanged',
          self._handle_services_changed,
        )
        self.call(
          'GetServices',
          reply_handler = self._handle_get_services_reply,
        )

    def _handle_get_technologies_reply(self, technology_tuples):
        self.technologies.add_initial_objects(
          [technology_tuple[0] for technology_tuple in technology_tuples]
        )

    def _handle_technologies_ready(self, technologies):
        self.maybe_ready()

    def _handle_get_services_reply(self, tuples):
        self.services.add_initial_objects(
          [path for (path, properties) in tuples]
        )

    def _handle_services_changed(self, tuples, removed):
        self.services.update_objects([path for (path, properties) in tuples])

    def _handle_services_ready(self, services):
        self.maybe_ready()

    def is_ready(self):
        if not super(ConnmanManager, self).is_ready():
            return False
        return (self.technologies.is_ready() and self.services.is_ready())

    def get_proxy_object(self):
        obj = super(ConnmanManager, self).get_proxy_object()
        return dbus.Interface(obj, '%s.Manager' % self.bus_name)


class ConnmanTechnology(DBusObjectWithConnmanProperties):
    def __init__(self, **kwargs):
        super(ConnmanTechnology, self).__init__(
          bus_name = 'net.connman',
          **kwargs
        )

    def get_proxy_object(self):
        obj = super(ConnmanTechnology, self).get_proxy_object()
        return dbus.Interface(obj, '%s.Technology' % self.bus_name)


class ConnmanService(DBusObjectWithConnmanProperties):
    def __init__(self, **kwargs):
        super(ConnmanService, self).__init__(
          bus_name = 'net.connman',
          **kwargs
        )

    def get_proxy_object(self):
        obj = super(ConnmanService, self).get_proxy_object()
        return dbus.Interface(obj, '%s.Service' % self.bus_name)
